// Copyright 2023 RozK
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
// OF SUCH DAMAGE.

using System;
using System.Collections.Generic;
using UnityEngine.Assertions;

namespace RK.MicroComponents {

public delegate void MicroFunc<_Param>(_Param param);

internal readonly struct MicroAction {

    internal readonly ulong Guid; // -> MicroActions
    internal readonly Delegate Func; // -> MicroActions
    private readonly object _param;

    internal static MicroAction Create<_Param>(ulong guid, MicroFunc<_Param> func) {
        return new MicroAction(guid, func, null);
    }

    internal static MicroAction Bind<_Param>(MicroAction action, _Param param) {
        return new MicroAction(action.Guid, action.Func, param);
    }

    internal static MicroAction Create<_Param>(ulong guid, MicroFunc<_Param> func, _Param param) {
        return new MicroAction(guid, func, param);
    }

    private MicroAction(ulong guid, Delegate func, object param) {
        Guid = guid;
        Func = func;
        _param = param;
    }

    internal void Invoke() {
        Func.DynamicInvoke(_param);
    }
}

public class MicroActions {

    internal List<MicroAction> ParallelActions; // -> MicroSystem
    internal List<MicroAction> SyncActions; // -> MicroSystem

    internal bool IsEmpty {
        get => (
            (ParallelActions is null || ParallelActions.Count == 0) &&
            (SyncActions is null || SyncActions.Count == 0)
        );
    }

    private void AddParallel(MicroAction action) {
        if (ParallelActions is null) {
            ParallelActions = new List<MicroAction>(1);
        }
        ParallelActions.Add(action);
    }

    internal void AddParallel<_Param>(MicroFunc<_Param> func) {
        MicroInstance instance = func.Target as MicroInstance;
        Assert.IsNotNull(instance, "Must be a MicroInstance.");
        if (instance.IsValid) {
            AddParallel(MicroAction.Create(instance.Guid, func));
        }
    }

    internal void AddParallel<_Param>(MicroFunc<_Param> func, _Param param) {
        MicroInstance instance = func.Target as MicroInstance;
        Assert.IsNotNull(instance, "Must be a MicroInstance.");
        if (instance.IsValid) {
            AddParallel(MicroAction.Create(instance.Guid, func, param));
        }
    }

    private void AddSync(MicroAction action) {
        if (SyncActions is null) {
            SyncActions = new List<MicroAction>(1);
        }
        SyncActions.Add(action);
    }

    internal void AddSync<_Param>(MicroFunc<_Param> func) {
        MicroInstance instance = func.Target as MicroInstance;
        Assert.IsNotNull(instance, "Must be a MicroInstance.");
        if (instance.IsValid) {
            AddSync(MicroAction.Create(instance.Guid, func));
        }
    }

    internal void AddSync<_Param>(MicroFunc<_Param> func, _Param param) {
        MicroInstance instance = func.Target as MicroInstance;
        Assert.IsNotNull(instance, "Must be a MicroInstance.");
        if (instance.IsValid) {
            AddSync(MicroAction.Create(instance.Guid, func, param));
        }
    }

    internal void AddActions(MicroActions actions) {
        if (actions.ParallelActions is not null && actions.ParallelActions.Count > 0) {
            if (ParallelActions is null) {
                ParallelActions = new List<MicroAction>(actions.ParallelActions.Count);
            }
            else {
                Preallocate(ParallelActions, actions.ParallelActions.Count);
            }
            ParallelActions.AddRange(actions.ParallelActions);
        }
        if (actions.SyncActions is not null && actions.SyncActions.Count > 0) {
            if (SyncActions is null) {
                SyncActions = new List<MicroAction>(actions.SyncActions.Count);
            }
            else {
                Preallocate(SyncActions, actions.SyncActions.Count);
            }
            SyncActions.AddRange(actions.SyncActions);
        }
    }

    internal void AddEvent<_Param>(MicroEvent<_Param> _event, _Param param) {
        _event.Mutex.Acquire();
        if (_event.ParallelActions is not null && _event.ParallelActions.Count > 0) {
            if (ParallelActions is null) {
                ParallelActions = new List<MicroAction>(_event.ParallelActions.Count);
            }
            else {
                Preallocate(ParallelActions, _event.ParallelActions.Count);
            }
            foreach (MicroAction action in _event.ParallelActions) {
                ParallelActions.Add(MicroAction.Bind(action, param));
            }
        }
        if (_event.SyncActions is not null && _event.SyncActions.Count > 0) {
            if (SyncActions is null) {
                SyncActions = new List<MicroAction>(_event.SyncActions.Count);
            }
            else {
                Preallocate(SyncActions, _event.SyncActions.Count);
            }
            foreach (MicroAction action in _event.SyncActions) {
                SyncActions.Add(MicroAction.Bind(action, param));
            }
        }
        _event.Mutex.Release();
    }

    private void Preallocate(List<MicroAction> actions, int count) {
        int capacity = actions.Count + count;
        if (actions.Capacity < capacity) {
            actions.Capacity = capacity;
        }
    }

    internal void RemoveParallel<_Param>(MicroFunc<_Param> func) {
        if (ParallelActions is not null) {
            ParallelActions.RemoveAll(
                action => action.Func.Equals(func)
            );
        }
    }

    internal void RemoveSync<_Param>(MicroFunc<_Param> func) {
        if (SyncActions is not null) {
            SyncActions.RemoveAll(
                action => action.Func.Equals(func)
            );
        }
    }

    internal void RemoveInstance(MicroInstance instance) {
        ulong guid = instance.Guid;
        if (ParallelActions is not null) {
            ParallelActions.RemoveAll(
                action => (action.Guid == guid)
            );
        }
        if (SyncActions is not null) {
            SyncActions.RemoveAll(
                action => (action.Guid == guid)
            );
        }
    }

    internal void Clear() {
        if (ParallelActions is not null) {
            ParallelActions.Clear();
        }
        if (SyncActions is not null) {
            SyncActions.Clear();
        }
    }
}

} // namespace RK.MicroComponents
