// Copyright 2023 RozK
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
// OF SUCH DAMAGE.

// TODO: Start/Stop by priority?

using System.Collections.Generic;
using UnityEngine;

namespace RK.MicroComponents {

[DefaultExecutionOrder(2)]
[DisallowMultipleComponent]
[AddComponentMenu("MicroComponents/MicroBehaviour")]

public sealed class MicroBehaviour : MonoBehaviour {

    [SerializeField]
    private List<MicroComponent> _microComponents;

    private bool _waitStart;
    private MicroInstance[] _instances;

    private void Awake() {
        if  (_microComponents is null) {
            Debug.LogError("[MicroComponents] Null component list.");
            Destroy(this);
            return;
        }
        for (int index = _microComponents.Count - 1; index >= 0; index--) {
            if (!_microComponents[index]) {
                Debug.LogWarning("[MicroComponents] Null component.");
                _microComponents.RemoveAt(index);
            }
        }
        if (_microComponents.Count == 0) {
            Debug.LogError("[MicroComponents] Empty component list.");
            Destroy(this);
            return;
        }
        _waitStart = enabled;
        _instances = new MicroInstance[_microComponents.Count];
        for (int index = 0; index < _microComponents.Count; index++) {
            MicroComponent component = _microComponents[index];
            _instances[index] = component.Manager.Allocate(this, component);
        }
    }

    private void Start() {
        if (_waitStart) {
            _waitStart = false;
            OnEnable();
        }
    }

    private void OnEnable() {
        if (!_waitStart) {
            for (int index = 0; index < _microComponents.Count; index++) {
                MicroInstance instance = _instances[index];
                if (instance.IsValid) {
                    MicroComponent component = _microComponents[index];
                    component.Manager.Start(instance);
                }
            }
        }
    }

    private void OnDisable() {
        for (int index = 0; index < _microComponents.Count; index++) {
            MicroInstance instance = _instances[index];
            if (instance.IsValid) {
                MicroComponent component = _microComponents[index];
                component.Manager.Stop(instance);
            }
        }
    }

    private void OnDestroy() {
        for (int index = 0; index < _microComponents.Count; index++) {
            MicroInstance instance = _instances[index];
            if (instance.IsValid) {
                MicroComponent component = _microComponents[index];
                component.Manager.Deallocate(instance);
            }
            _instances[index] = null;
        }
        _instances = null;
    }

    internal _Instance GetInstance<_Component, _Instance>()
        where _Component : MicroComponent
        where _Instance : MicroInstance<_Component, _Instance>, new() {
        for (int index = 0; index < _microComponents.Count; index++) {
            MicroComponent component = _microComponents[index];
            if (component is _Component) {
                return _instances[index] as _Instance;
            }
        }
        return null;
    }

    internal _Instance GetInstance<_Component, _Instance>(_Component _component)
        where _Component : MicroComponent
        where _Instance : MicroInstance<_Component, _Instance>, new() {
        for (int index = 0; index < _microComponents.Count; index++) {
            MicroComponent component = _microComponents[index];
            if (component == _component) {
                return _instances[index] as _Instance;
            }
        }
        return null;
    }
}

} // namespace RK.MicroComponents
