// Copyright 2023 RozK
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
// OF SUCH DAMAGE.

using System;
using System.Collections.Generic;
using UnityEngine;

namespace RK.MicroComponents {

[Flags]
public enum MicroUpdates {
    None = 0,
    ParallelFixedUpdate = 1,
    SyncFixedUpdate = 2,
    ParallelUpdate = 4,
    SyncUpdate = 8,
    ParallelLateUpdate = 16,
    SyncLateUpdate = 32
}

public abstract class MicroInstance {

    private static ulong _nextGuid = 1;

    internal ulong Guid; // -> MicroActions
    internal int InstanceId; // <- MicroManager
    internal MicroActions Actions; // <- MicroManager

    private MicroBehaviour _behaviour;
    public MicroBehaviour Behaviour { get => _behaviour; }

    internal bool IsValid { get => (Guid > 0); }

    internal protected MicroInstance() {
        Guid = 0;
        InstanceId = -1;
    }

    internal void Allocate(int instanceId, MicroBehaviour behaviour) {
        Guid = _nextGuid++;
        InstanceId = instanceId;
        _behaviour = behaviour;
    }

    internal void Deallocate() {
        Guid = 0;
        InstanceId = -1;
        _behaviour = null;
    }

    internal protected virtual void Initialize() {
    }

    internal protected virtual void Terminate() {
    }

    internal protected virtual void Start() {
    }

    internal protected virtual void Stop() {
    }

    internal protected virtual void ParallelFixedUpdate() {
    }

    internal protected virtual void SyncFixedUpdate() {
    }

    internal protected virtual void ParallelUpdate() {
    }

    internal protected virtual void SyncUpdate() {
    }

    internal protected virtual void ParallelLateUpdate() {
    }

    internal protected virtual void SyncLateUpdate() {
    }

    protected void ParallelCall<_Param>(MicroFunc<_Param> func, _Param param) {
        Actions.AddParallel(func, param);
    }

    protected void SyncCall<_Param>(MicroFunc<_Param> func, _Param param) {
        Actions.AddSync(func, param);
    }

    protected void InvokeEvent<_Param>(MicroEvent<_Param> _event, _Param param) {
        Actions.AddEvent(_event, param);
    }
}

public abstract class MicroInstance<_Component, _Instance> : MicroInstance
    where _Component : MicroComponent
    where _Instance : MicroInstance<_Component, _Instance>, new() {

    private static MicroConfiguration _configuration;
    internal MicroConfiguration Configuration { get => _configuration; }

    protected static void Configure(
        MicroUpdates updates,
        int parallelBatchSize = 1,
        int poolMinimumSize = 0,
        int poolMaximumSize = 0,
        int poolGranularity = 1) {
        _configuration.Updates = updates;
        _configuration.ParallelBatchSize = (parallelBatchSize > 0) ? parallelBatchSize : 1;
        _configuration.PoolMinimumSize = (poolMinimumSize >= 0) ? poolMinimumSize : 0;
        _configuration.PoolMaximumSize = (poolMaximumSize >= 0) ? poolMaximumSize : 0;
        _configuration.PoolGranularity = (poolGranularity > 0) ? poolGranularity : 1;
        if (_configuration.PoolMaximumSize > 0 &&
            _configuration.PoolMaximumSize < _configuration.PoolMinimumSize) {
            Debug.LogError("[MicroComponents] Pool maximum size must be >= minimum size.");
            _configuration.PoolMaximumSize = _configuration.PoolMinimumSize;
        }
    }

    public static void UpdateAfter<_Dependency>()
        where _Dependency : MicroInstance {
        if (_configuration.Dependencies is null) {
            _configuration.Dependencies = new HashSet<Type>(1);
        }
        _configuration.Dependencies.Add(typeof(_Dependency));
    }


    public static Reference From(MicroBehaviour behaviour) {
        return behaviour.GetInstance<_Component, _Instance>();
    }

    public static Reference From(MicroBehaviour behaviour, _Component component) {
        return behaviour.GetInstance<_Component, _Instance>(component);
    }

    private _Component _component;
    public _Component Component { get => _component; }

    internal void Allocate(int instanceId, MicroBehaviour behaviour, _Component component) {
        base.Allocate(instanceId, behaviour);
        _component = component;
    }

    internal new void Deallocate() {
        base.Deallocate();
        _component = null;
    }

    public struct Reference {

        private ulong _guid;
        private _Instance _instance;

        public _Instance Instance {
            get {
                if (_instance is not null && _instance.Guid != _guid) {
                    _guid = 0;
                    _instance = null;
                }
                return _instance;
            }
            set {
                if (value is not null && value.IsValid) {
                    _guid = value.Guid;
                    _instance = value;
                }
                else {
                    _guid = 0;
                    _instance = null;
                }
            }
        }

        public static implicit operator Reference(_Instance instance) {
            Reference reference = new Reference();
            reference.Instance = instance;
            return reference;
        }

        public static implicit operator _Instance(Reference reference) {
            return reference.Instance;
        }
    }
}

} // namespace RK.MicroComponents
