// Copyright 2023 RozK
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
// OF SUCH DAMAGE.

using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Threading.Tasks;
using UnityEngine;

namespace RK.MicroComponents {

internal struct MicroConfiguration {

    internal MicroUpdates Updates;
    internal int ParallelBatchSize;
    internal int PoolMinimumSize;
    internal int PoolMaximumSize;
    internal int PoolGranularity;
    internal HashSet<Type> Dependencies;
}

internal abstract class MicroManager {

    internal int Priority; // <- MicroSystem
    internal MicroConfiguration Configuration; // -> MicroSystem
    internal ConcurrentQueue<MicroActions> ThreadsActions; // <- MicroSystem
    protected int _startedCount;
    protected int _totalCount;

    internal abstract Type InstanceType { get; }

    internal abstract MicroInstance Allocate(MicroBehaviour behaviour, MicroComponent component);
    internal abstract void Deallocate(MicroInstance instance);
    internal abstract void Start(MicroInstance instance);
    internal abstract void Stop(MicroInstance instance);

    internal abstract void ParallelFixedUpdate();
    internal abstract void SyncFixedUpdate();
    internal abstract void ParallelUpdate();
    internal abstract void SyncUpdate();
    internal abstract void ParallelLateUpdate();
    internal abstract void SyncLateUpdate();
}

internal sealed class MicroManager<_Component, _Instance> : MicroManager
    where _Component : MicroComponent
    where _Instance : MicroInstance<_Component, _Instance>, new() {

    private List<_Instance> _instances;

    internal sealed override Type InstanceType { get => typeof(_Instance); }

    internal MicroManager() {
        _Instance instance = new _Instance();
        Configuration = instance.Configuration;
        _startedCount = 0;
        _totalCount = 0;
        int capacity = Configuration.PoolGranularity;
        if (Configuration.PoolMinimumSize > capacity) {
            int top = Configuration.PoolMinimumSize + Configuration.PoolGranularity - 1;
            capacity = (top / Configuration.PoolGranularity) * Configuration.PoolGranularity;
        }
        _instances = new List<_Instance>(capacity);
        if (Configuration.PoolMinimumSize > 0) {
            _instances.Add(instance);
            for (int index = 1; index < Configuration.PoolMinimumSize; index++) {
                _instances.Add(new _Instance());
            }
        }
        MicroSystem.RegisterManager(this);
    }

    internal sealed override MicroInstance Allocate(MicroBehaviour behaviour, MicroComponent _component) {
        if (Configuration.PoolMaximumSize > 0 && _totalCount == Configuration.PoolMaximumSize) {
            Debug.LogError("[MicroComponents] Pool reached maximum size.");
            return null;
        }
        _Component component = _component as _Component;
        if (!component) {
            Debug.LogError("[MicroComponents] Null component.");
            return null;
        }
        _Instance instance;
        if (_totalCount < _instances.Count) {
            instance = _instances[_totalCount];
        }
        else {
            if (_totalCount == _instances.Capacity) {
                _instances.Capacity = _totalCount + Configuration.PoolGranularity;
            }
            instance = new _Instance();
            _instances.Add(instance);
        }
        instance.Allocate(_totalCount++, behaviour, component);
        instance.Initialize();
        return instance;
    }

    internal sealed override void Deallocate(MicroInstance _instance) {
        int instanceId = _instance.InstanceId;
        if (instanceId >= _startedCount && instanceId < _totalCount) {
            _Instance instance = _instances[instanceId];
            Swap(instanceId, --_totalCount);
            instance.Terminate();
            instance.Deallocate();
            int count = _instances.Count;
            if (count > Configuration.PoolMinimumSize) {
                _instances.RemoveAt(--count);
            }
            int capacity = _instances.Capacity;
            if ((capacity - count) >= Configuration.PoolGranularity) {
                _instances.Capacity = capacity - Configuration.PoolGranularity;
            }
        }
        else {
            Debug.LogError("[MicroComponents] InstanceId is invalid.");
        }
    }

    internal sealed override void Start(MicroInstance instance) {
        int instanceId = instance.InstanceId;
        if (instanceId >= _startedCount && instanceId < _totalCount) {
            Swap(instanceId, _startedCount++);
            instance.Start();
        }
        else {
            Debug.LogError("[MicroComponents] Instance is invalid or already started.");
        }
    }

    internal sealed override void Stop(MicroInstance instance) {
        int instanceId = instance.InstanceId;
        if (instanceId >= 0 && instanceId < _startedCount) {
            instance.Stop();
            Swap(instanceId, --_startedCount);
        }
        else {
            Debug.LogError("[MicroComponents] Instance is invalid or already stopped.");
        }
    }

    private void Swap(int instanceId, int otherInstanceId) {
        if (instanceId != otherInstanceId) {
            _Instance instance = _instances[instanceId];
            _Instance otherInstance = _instances[otherInstanceId];
            _instances[instanceId] = otherInstance;
            _instances[otherInstanceId] = instance;
            otherInstance.InstanceId = instanceId;
            instance.InstanceId = otherInstanceId;
        }
    }

    internal sealed override void ParallelFixedUpdate() {
        ParallelInvokeUpdate(instance => instance.ParallelFixedUpdate());
    }

    internal sealed override void SyncFixedUpdate() {
        SyncInvokeUpdate(instance => instance.SyncFixedUpdate());
    }

    internal sealed override void ParallelUpdate() {
        ParallelInvokeUpdate(instance => instance.ParallelUpdate());
    }

    internal sealed override void SyncUpdate() {
        SyncInvokeUpdate(instance => instance.SyncUpdate());
    }

    internal sealed override void ParallelLateUpdate() {
        ParallelInvokeUpdate(instance => instance.ParallelLateUpdate());
    }

    internal sealed override void SyncLateUpdate() {
        SyncInvokeUpdate(instance => instance.SyncLateUpdate());
    }

    private void InvokeUpdate(int instanceId, Action<_Instance> invokeUpdate, MicroActions actions) {
        _Instance instance = _instances[instanceId];
        instance.Actions = actions;
        invokeUpdate(instance);
        instance.Actions = null;
    }

    private void InvokeUpdate(int firstId, int lastId, Action<_Instance> invokeUpdate, MicroActions actions) {
        for (int instanceId = firstId; instanceId < lastId; instanceId++) {
            InvokeUpdate(instanceId, invokeUpdate, actions);
        }
    }

    private void ParallelInvokeUpdate(Action<_Instance> invokeUpdate) {
        if (_startedCount > 0) {
            int batchSize = Configuration.ParallelBatchSize;
            if (batchSize == 1) {
                Parallel.For(0, _startedCount, MicroSystem.ParallelOptions,
                    () => {
                        ThreadsActions.TryDequeue(out MicroActions actions);
                        return actions;
                    },
                    (instanceId, _, actions) => {
                        InvokeUpdate(instanceId, invokeUpdate, actions);
                        return actions;
                    },
                    actions => {
                        ThreadsActions.Enqueue(actions);
                    }
                );
            }
            else {
                int batchCount = (_startedCount + batchSize - 1) / batchSize;
                Parallel.For(0, batchCount, MicroSystem.ParallelOptions,
                    () => {
                        if (!ThreadsActions.TryDequeue(out MicroActions actions)) {
                            Debug.LogError("[MicroComponents] Too many threads.");
                            actions = new MicroActions();
                        }
                        return actions;
                    },
                    (batchIndex, _, actions) => {
                        int firstId = batchIndex * batchSize;
                        int lastId = firstId + batchSize;
                        if (lastId > _startedCount) {
                            lastId = _startedCount;
                        }
                        InvokeUpdate(firstId, lastId, invokeUpdate, actions);
                        return actions;
                    },
                    actions => {
                        ThreadsActions.Enqueue(actions);
                    }
                );
            }
        }
    }

    private void SyncInvokeUpdate(Action<_Instance> invokeUpdate) {
        ThreadsActions.TryDequeue(out MicroActions actions);
        InvokeUpdate(0, _startedCount, invokeUpdate, actions);
        ThreadsActions.Enqueue(actions);
    }
}

} // namespace RK.MicroComponents
