// Copyright 2023 RozK
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
// OF SUCH DAMAGE.

// TODO: parallel/sync calls by priority?

using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Threading.Tasks;
using System.Linq;
using UnityEngine;

namespace RK.MicroComponents {

[DefaultExecutionOrder(1)]
[DisallowMultipleComponent]
[AddComponentMenu("MicroComponents/MicroSystem")]

public sealed class MicroSystem : MonoBehaviour {

    public class _TimeCache {
        internal double _fixedTime;
        internal double _time;
        internal float _fixedDeltaTime;
        internal float _deltaTime;

        public double FixedTime { get => _fixedTime; }
        public double Time { get => _time; }
        public float FixedDeltaTime { get => _fixedDeltaTime; }
        public float DeltaTime { get => _deltaTime; }

        internal void Update() {
            _timeCache._fixedTime = UnityEngine.Time.fixedTimeAsDouble;
            _timeCache._time = UnityEngine.Time.timeAsDouble;
            _timeCache._fixedDeltaTime = UnityEngine.Time.fixedDeltaTime;
            _timeCache._deltaTime = UnityEngine.Time.deltaTime;
        }
    }

    internal static _TimeCache _timeCache = new _TimeCache();
    public static _TimeCache TimeCache { get => _timeCache; }

    internal static readonly ParallelOptions ParallelOptions = new ParallelOptions {
        MaxDegreeOfParallelism = Environment.ProcessorCount
    };

    private static MicroSystem _instance;
    private static List<MicroManager> _managers;

    internal static void RegisterManager(MicroManager manager) {
        if (_managers is null) {
            _managers = new List<MicroManager>();
        }
        _managers.Add(manager);
    }

    private Dictionary<Type, MicroManager> _managersByType;
    private List<MicroManager> _parallelFixedUpdateManagers;
    private List<MicroManager> _syncFixedUpdateManagers;
    private List<MicroManager> _parallelUpdateManagers;
    private List<MicroManager> _syncUpdateManagers;
    private List<MicroManager> _parallelLateUpdateManagers;
    private List<MicroManager> _syncLateUpdateManagers;
    private ConcurrentQueue<MicroActions> _threadsActions;
    private MicroActions _actions;

    private void EvaluatePriorities() {
        foreach (MicroManager manager in _managers) {
            FixCyclicDependencies(manager);
            manager.Priority = -1;
        }
        foreach (MicroManager manager in _managers) {
            if (manager.Priority < 0) {
                EvaluatePriority(manager);
            }
        }
    }

    private void EvaluatePriority(MicroManager manager) {
        manager.Priority = 0;
        HashSet<Type> dependencies = manager.Configuration.Dependencies;
        if (dependencies is not null) {
            foreach (Type dependencyType in dependencies) {
                if (_managersByType.TryGetValue(dependencyType, out MicroManager dependency)) {
                    if (dependency.Priority < 0) {
                        EvaluatePriority(dependency);
                    }
                    if (dependency.Priority > manager.Priority) {
                        manager.Priority = dependency.Priority;
                    }
                }
            }
            manager.Priority += 1;
        }
    }

    private void FixCyclicDependencies(MicroManager manager) {
        Type instanceType = manager.InstanceType;
        List<MicroManager> cyclicDependencies = new List<MicroManager>();
        FindDependenciesWithDependency(manager, instanceType, cyclicDependencies);
        foreach (MicroManager dependency in cyclicDependencies) {
            Type dependencyType = dependency.InstanceType;
            Debug.LogError(
                $"[MicroComponents] Cyclic dependency: {dependencyType} -> {instanceType}");
            dependency.Configuration.Dependencies.Remove(instanceType);
        }
    }

    private void FindDependenciesWithDependency(MicroManager manager, Type type, List<MicroManager> found) {
        HashSet<Type> dependencies = manager.Configuration.Dependencies;
        if (dependencies is not null) {
            if (dependencies.Contains(type)) {
                found.Add(manager);
            }
            foreach (Type dependencyType in dependencies) {
                if (_managersByType.TryGetValue(dependencyType, out MicroManager dependency)) {
                    FindDependenciesWithDependency(dependency, type, found);
                }
            }
        }
    }

    private void Awake() {
        if (_instance) {
            Debug.LogError(
                "[MicroComponents] Only one MicroSystem is allowed at any given time.");
            Destroy(this);
            return;
        }
        _timeCache.Update();
        _instance = this;
        _managersByType = new Dictionary<Type, MicroManager>();
        foreach (MicroManager manager in _managers) {
            _managersByType.Add(manager.InstanceType, manager);
        }
        EvaluatePriorities();
        _managers.Sort((a, b) => a.Priority.CompareTo(b.Priority));
        _parallelFixedUpdateManagers = new List<MicroManager>();
        _syncFixedUpdateManagers = new List<MicroManager>();
        _parallelUpdateManagers = new List<MicroManager>();
        _syncUpdateManagers = new List<MicroManager>();
        _parallelLateUpdateManagers = new List<MicroManager>();
        _syncLateUpdateManagers = new List<MicroManager>();
        foreach (MicroManager manager in _managers) {
            MicroUpdates updates = manager.Configuration.Updates;
            if ((updates & MicroUpdates.ParallelFixedUpdate) != MicroUpdates.None) {
                _parallelFixedUpdateManagers.Add(manager);
            }
            if ((updates & MicroUpdates.SyncFixedUpdate) != MicroUpdates.None) {
                _syncFixedUpdateManagers.Add(manager);
            }
            if ((updates & MicroUpdates.ParallelUpdate) != MicroUpdates.None) {
                _parallelUpdateManagers.Add(manager);
            }
            if ((updates & MicroUpdates.SyncUpdate) != MicroUpdates.None) {
                _syncUpdateManagers.Add(manager);
            }
            if ((updates & MicroUpdates.ParallelLateUpdate) != MicroUpdates.None) {
                _parallelLateUpdateManagers.Add(manager);
            }
            if ((updates & MicroUpdates.SyncLateUpdate) != MicroUpdates.None) {
                _syncLateUpdateManagers.Add(manager);
            }
        }
        _threadsActions = new ConcurrentQueue<MicroActions>();
        for (int index = 0; index < ParallelOptions.MaxDegreeOfParallelism; index++) {
            _threadsActions.Enqueue(new MicroActions());
        }
        _actions = new MicroActions();
    }

    private void OnDestroy() {
        _instance = null;
    }

    private void FixedUpdate() {
        _timeCache.Update();
        InvokeUpdate(_parallelFixedUpdateManagers, manager => manager.ParallelFixedUpdate());
        InvokeUpdate(_syncFixedUpdateManagers, manager => manager.SyncFixedUpdate());
    }

    private void Update() {
        _timeCache.Update();
        InvokeUpdate(_parallelUpdateManagers, manager => manager.ParallelUpdate());
        InvokeUpdate(_syncUpdateManagers, manager => manager.SyncUpdate());
    }

    private void LateUpdate() {
        _timeCache.Update();
        InvokeUpdate(_parallelLateUpdateManagers, manager => manager.ParallelLateUpdate());
        InvokeUpdate(_syncLateUpdateManagers, manager => manager.SyncLateUpdate());
    }

    private void InvokeUpdate(List<MicroManager> managers, Action<MicroManager> invokeUpdate) {
        if (managers.Count > 0) {
            foreach (MicroManager manager in managers) {
                manager.ThreadsActions = _threadsActions;
                invokeUpdate(manager);
                manager.ThreadsActions = null;
            }
            foreach (MicroActions actions in _threadsActions) {
                _actions.AddActions(actions);
                actions.Clear();
            }
            if (_actions.ParallelActions is not null && _actions.ParallelActions.Count > 0) {
                ParallelQuery<MicroAction> actions = _actions.ParallelActions.AsParallel()
                    .WithExecutionMode(ParallelExecutionMode.ForceParallelism)
                    .WithDegreeOfParallelism(Environment.ProcessorCount);
                (from action in actions group action by action.Guid).ForAll(
                    group => {
                        foreach (MicroAction action in group) {
                            action.Invoke();
                        }
                    }
                );
                _actions.ParallelActions.Clear();
            }
            if (_actions.SyncActions is not null && _actions.SyncActions.Count > 0) {
                foreach(MicroAction action in _actions.SyncActions) {
                    action.Invoke();
                }
                _actions.SyncActions.Clear();
            }
        }
    }
}

} // namespace RK.MicroComponents
